#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module contains the Just-In-Time compiler tools.
The main function is jit(form, options), see its documentation.
"""

# Copyright (C) 2008-2009 Martin Sandve Alnes and Simula Resarch Laboratory
#
# This file is part of SyFi.
#
# SyFi is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#
# SyFi is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with SyFi. If not, see <http://www.gnu.org/licenses/>.
#
# Modified by Kent-Andre Mardal, 2010.
#
# First added:  2008-09-04
# Last changed: 2009-04-23

import shutil
import os
import hashlib
import instant
import ufc_utils
import ufl
import ufl.algorithms
from ufl.algorithms import preprocess

from sfc.common import version
from sfc.common import default_parameters
from sfc.common import ParameterDict
from sfc.common import get_abs_ufc_h_path
from sfc.common import sfc_debug, sfc_info, sfc_warning, sfc_error, sfc_assert, write_file
from sfc.common.names import short_name, form_classname, dof_map_classname, finite_element_classname, base_element_classname
from sfc.codegeneration import generate_code, compiler_input

def cache_signature(input, options):
    "A signature string uniquely identifying jit input."

    # FIXME, need to use i.form_data() for memory caching  TODO: What does this mean?
    if not isinstance(input, list):
        input = [input]
    newinput = []
    for i in input:
        if isinstance(i, ufl.form.Form):
            i = i.signature()
            #i = i.compute_form_data().preprocessed_form.signature() # TODO: Something like this?
        newinput.append(i)
    return "sfc.jit(%r, %r) # SFC version %r" % (newinput, options, version)

def compile_module(module_dir, hfilenames, cfilenames, signature, options):
    """Assuming an existing directory module_dir with source files
    hfilenames, cfilenames, create a python extension module and compile it."""

    orig_dir = os.getcwd()
    try:
        # headers to #include in swig file
        system_headers = ["iostream", "stdexcept", "cmath", "ufc.h", "tr1/memory"]

        cppargs = options.compilation.cppargs
        if options.compilation.enable_debug_code:
            cppargs.append("-DSFCDEBUG")

        module_name = module_dir if signature is None else None

        module = ufc_utils.build_ufc_module(h_files = hfilenames,
                     source_directory   = module_dir,
                     sources            = cfilenames,
                     system_headers     = system_headers,
                     cppargs            = cppargs,
                     signature          = signature,
                     cache_dir          = options.compilation.cache_dir,
                     modulename         = module_name,
                     generate_interface = options.compilation.generate_interface,
                     generate_setup     = options.compilation.generate_setup)
        sfc_info("Successfully compiled module in '%s'." % module_dir)
        return module
        # ...
    finally:
        os.chdir(orig_dir)

def jit(input, parameters=None, common_cell=None):
    ufl_elements, formdatas = compiler_input(input, common_cell=common_cell)

    # Merge parameters with defaults
    if parameters is None:
        parameters = default_parameters()
    if isinstance(parameters, dict) and not isinstance(parameters, ParameterDict):
        o = parameters
        parameters = default_parameters()
        parameters.update(o)
    
    # Default prefix shows where the module was compiled
    prefix = parameters.code.prefix
    if not prefix:
        prefix = "sfc_jit"

    # Build module signature and classnames from input
    signature = cache_signature(input, parameters)
    formclassnames = [form_classname(fd.preprocessed_form, parameters) for fd in formdatas]
#    formclassnames = [form_classname(fd.form, parameters) for fd in formdatas]
    dmclassnames = [dof_map_classname(e) for e in ufl_elements]
    feclassnames = [finite_element_classname(e) for e in ufl_elements]

    # Placeholder for the up and coming compiled module
    module = None


    # Hacky hook for profiling without the C++ compilation
    if parameters.compilation.skip:
        pass
    else:
        # Possibly force recompilation
        if parameters.compilation.overwrite_cache:
            sfc_warning("Complete implementation of overwrite_cache not done, but skipping cache lookup before code generation.")
            # TODO: delete existing module with signature or module_name
        else:
            # Check cache before generating code
            module = instant.import_module(signature)
            # Construct ufc::form object
            if module:
                sfc_info("Found module in cache.")
    
    if module is None:
        # Store original path, to restore later
        orig_dir = os.getcwd()

        # Build module name, then shorten if it's too long
        module_dir_name = "_".join(["_".join(formclassnames), "_".join(dmclassnames), "_".join(feclassnames)])
        module_dir_name = short_name(prefix, module_dir_name)
        
        # Make sure the module directory is there and empty
        module_dir = os.path.abspath(module_dir_name)
        if os.path.exists(module_dir):
            sfc_warning("Possibly overwriting files in existing module directory '%s'." % module_dir)
        else:
            os.mkdir(module_dir)
        
        # Enter module directory and generate code!
        try:
            os.chdir(module_dir)
            
            # Write parameters, signature and form representations to files beside code
            write_file("parameters.repr", repr(parameters))
            write_file("signature", signature)
            for fd in formdatas:
                write_file("%s.repr" % short_name("form_", fd.name), repr(fd.preprocessed_form))
            for ue in ufl_elements:
                write_file("%s.repr" % short_name("element_", base_element_classname(ue)), repr(ue))

            # Generate code!
            hfilenames, cfilenames = generate_code(ufl_elements + formdatas, None,
                                                   parameters, common_cell=common_cell)
            
            # Hook for profiling without the C++ compilation
            if parameters.compilation.skip:
                return None

        finally:
            # Restore original path
            os.chdir(orig_dir)
        
        # Invoke Instant to compile generated files as a Python extension module
        module = compile_module(module_dir, hfilenames, cfilenames, signature, parameters)
        
        # Remove temporary module directory if code was copied to cache
        placed_in_cache = False # TODO
        if placed_in_cache:
            sfc_info("Module was placed in cache, deleting module directory '%s'." % module_dir)
            shutil.rmtree(module_dir)
        
        if module is None:
            sfc_error("Failed to load compiled module from cache.")
    
    # If we get here, 'module' should be our compiled module.
    
    # TODO: Handle mixed list of forms and elements?

    # Got forms, return forms
    if formdatas:
        prefix = "sfc dummy prefix for pydolfin which I have no idea what is used for"
        ufc_forms = [getattr(module, name)() for name in formclassnames]
        return ufc_forms[0], module, formdatas[0], prefix

    # Got elements, return elements
    elif ufl_elements:
        ufc_elements = [(getattr(module, fe)(), getattr(module, dm)()) for (fe, dm) in zip(feclassnames, dmclassnames)]
        return ufc_elements[0]
    
    msg = "Nothing to return, this shouldn't happen.\ninput =\n%r" % repr(input)
    sfc_error(msg)


def jitf(input, objects, parameters = None):
    """Generate code from input and parameters.
    
    @param input:
        TODO
    @param parameters:
        TODO
    @return:
        FIXME: Agree on return values with FFC! Currently FFC returns (ufc_form, module, formdata), and dolfin assumes this
    """

    # To handle all input uniformly
    list_input = isinstance(input, list)
    ufl_elements, formdatas = compiler_input(input, objects=objects)

    # Merge parameters with defaults
    if parameters is None:
        parameters = default_parameters()
    if isinstance(parameters, dict) and not isinstance(parameters, ParameterDict):
        o = parameters
        parameters = default_parameters()
        parameters.update(o)
    
    # Default prefix shows where the module was compiled
    prefix = parameters.code.prefix
    if not prefix:
        prefix = "sfc_jit"

    # Build module signature and classnames from input
    signature = cache_signature(input, parameters)
    formclassnames = [form_classname(fd.preprocessed_form, parameters) for fd in formdatas]
    dmclassnames = [dof_map_classname(e) for e in ufl_elements]
    feclassnames = [finite_element_classname(e) for e in ufl_elements]

    # Placeholder for the up and coming compiled module
    module = None

    # Hacky hook for profiling without the C++ compilation
    if parameters.compilation.skip:
        pass
    else:
        # Possibly force recompilation
        if parameters.compilation.overwrite_cache:
            sfc_warning("Complete implementation of overwrite_cache not done, but skipping cache lookup before code generation.")
            # TODO: delete existing module with signature or module_name
        else:
            # Check cache before generating code
            module = instant.import_module(signature)
            # Construct ufc::form object
            if module:
                sfc_info("Found module in cache.")
    
    if module is None:
        # Store original path, to restore later
        orig_dir = os.getcwd()

        # Build module name, then shorten if it's too long
        module_dir_name = "_".join(["_".join(formclassnames), "_".join(dmclassnames), "_".join(feclassnames)])
        module_dir_name = short_name(prefix, module_dir_name)
        
        # Make sure the module directory is there and empty
        module_dir = os.path.abspath(module_dir_name)
        if os.path.exists(module_dir):
            sfc_warning("Possibly overwriting files in existing module directory '%s'." % module_dir)
        else:
            os.mkdir(module_dir)
        
        # Enter module directory and generate code!
        try:
            os.chdir(module_dir)
            
            # Write parameters, signature and form representations to files beside code
            write_file("parameters.repr", repr(parameters))
            write_file("signature", signature)
            for fd in formdatas:
                write_file("%s.repr" % short_name("form_", fd.name), repr(fd.preprocessed_form))
            for ue in ufl_elements:
                write_file("%s.repr" % short_name("element_", base_element_classname(ue)), repr(ue))
            
            # Generate code!
            hfilenames, cfilenames = generate_code(ufl_elements + formdatas, objects,
                                                   parameters)
            
            # Hook for profiling without the C++ compilation
            if parameters.compilation.skip:
                return None

        finally:
            # Restore original path
            os.chdir(orig_dir)
        
        # Invoke Instant to compile generated files as a Python extension module
        module = compile_module(module_dir, hfilenames, cfilenames, signature, parameters)
        
        # Remove temporary module directory if code was copied to cache
        placed_in_cache = False # TODO
        if placed_in_cache:
            sfc_info("Module was placed in cache, deleting module directory '%s'." % module_dir)
            shutil.rmtree(module_dir)
        
        if module is None:
            sfc_error("Failed to load compiled module from cache.")
    
    # If we get here, 'module' should be our compiled module.
    
    # TODO: Handle mixed list of forms and elements?
    
    # Got forms, return forms
    if formdatas:
        prefix = "sfc dummy prefix for pydolfin which I have no idea what is used for"
        ufc_forms = [getattr(module, name) for name in formclassnames]
        if list_input:
            return ufc_forms, module, formdatas, [prefix]*len(ufc_forms)
        else:
            return ufc_forms[0], module, formdatas[0], prefix

    # Got elements, return elements
    elif ufl_elements:
        ufc_elements = [(getattr(module, fe)(), getattr(module, dm)()) for (fe, dm) in zip(feclassnames, dmclassnames)]
        if list_input:
            return ufc_elements
        else:
            return ufc_elements[0]
    
    msg = "Nothing to return, this shouldn't happen.\ninput =\n%r" % repr(input)
    sfc_error(msg)
